@@ -1,6 +1,6 @@
 ;;; ebib-notes.el --- Part of Ebib, a BibTeX database manager  -*- lexical-binding: t -*-
 
-;; Copyright (c) 2003-2018 Joost Kremers
+;; Copyright (c) 2003-2019 Joost Kremers
 ;; All rights reserved.
 
 ;; Author: Joost Kremers <joostkremers@fastmail.fm>
@@ -170,6 +170,15 @@ string where point should be located."
             (ebib--notes-locate-note key)))
     (file-readable-p (ebib--create-notes-file-name key))))
 
+(defun ebib-notes-display-note-symbol (_field key _db)
+  "Return the note symbol for displaying if a note exists for KEY."
+  (if (ebib--notes-exists-note key)
+      (propertize ebib-notes-symbol
+                  'face '(:height 0.8 :inherit link)
+                  'mouse-face 'highlight)
+    (propertize (make-string (string-width ebib-notes-symbol) ?\s)
+                'face '(:height 0.8))))
+
 ;;; One file per note.
 
 (defun ebib--create-notes-file-name (key)
@@ -226,7 +235,7 @@ This function also runs `ebib-notes-search-note-before-hook'."
   (run-hooks 'ebib-notes-search-note-before-hook)
   (save-excursion
     (goto-char (point-min))
-    (search-forward (funcall (cdr (assoc ?K ebib-notes-template-specifiers)) key nil) nil t)))
+    (re-search-forward (concat (regexp-quote (funcall (cdr (assoc ?K ebib-notes-template-specifiers)) key nil)) "$") nil t)))
 
 (defun ebib--notes-open-common-notes-file (key db)
   "Open the notes file for entry KEY in DB or create a new note."
