@@ -1,6 +1,6 @@
 ;;; ebib-utils.el --- Part of Ebib, a BibTeX database manager  -*- lexical-binding: t -*-
 
-;; Copyright (c) 2003-2018 Joost Kremers
+;; Copyright (c) 2003-2019 Joost Kremers
 ;; All rights reserved.
 
 ;; Author: Joost Kremers <joostkremers@fastmail.fm>
@@ -37,8 +37,9 @@
 ;;; Code:
 
 (require 'cl-lib)
-(require 'dash)
+(require 'seq)
 (require 'bibtex)
+(require 'format-spec)
 (require 'ebib-db)
 
 ;; Make a bunch of variables obsolete.
@@ -255,6 +256,15 @@ not change the default sort."
                         (integer :tag "Width")
                         (boolean :tag "Sort"))))
 
+(defcustom ebib-index-default-sort nil
+  "Default sort field and direction."
+  :group 'ebib
+  :type '(choice (const :tag "Use First Index Column" nil)
+                 (cons :tag "User-Defined Sort"
+                       (string :tag "Sort Field")
+                       (choice (const :tag "Ascending Sort" ascend)
+                               (const :tag "Descending Sort" descend)))))
+
 (defcustom ebib-index-column-separator "  "
   "Separator between columns in the index buffer."
   :group 'ebib
@@ -263,7 +273,7 @@ not change the default sort."
 (defcustom ebib-field-transformation-functions '(("Title" . ebib-clean-TeX-markup)
                                              ("Doi" . ebib-display-www-link)
                                              ("Url" . ebib-display-www-link)
-                                             ("Note" . ebib-display-note-symbol))
+                                             ("Note" . ebib-notes-display-note-symbol))
   "Functions transforming field contents to appropriate forms.
 Each function should accept three arguments, the field to be
 displayed, the key of the entry being displayed, and the database
@@ -378,7 +388,7 @@ contains the entry."
   :group 'ebib
   :type '(choice (function-item :tag "Author/Year" ebib-author-year-description)
                  (function-item :tag "Title" ebib-title-description)
-                 (function :tag "Custom function")))
+                 (function :tag "Custom Function")))
 
 (defun ebib-author-year-description (key db)
   "Provide an author/year description for an Org Ebib link.
@@ -399,6 +409,48 @@ database that contains the entry."
   :group 'ebib
   :type '(function :tag "Mode function"))
 
+(defcustom ebib-multiline-display-function 'ebib-multiline-display-as-is
+  "The way in which multiline field values are shown in the index buffer.
+By default, the value is shown as-is, but never more than
+`ebib-multiline-display-max-lines' lines.  The option \"First
+Paragraph\" displays the first paragraph (also with a maximum of
+`ebib-multiline-display-max-lines' lines), but in addition
+refills the text, which is useful if you use long lines in your
+multiline values.
+
+It is possible to provide a custom function for this option.
+This function should take a (multiline) string as argument and
+return a list of lines to be displayed."
+  :group 'ebib
+  :type '(choice (function-item :tag "Show Value As-Is" ebib-multiline-display-as-is)
+                 (function-item :tag "Show First Paragraph" ebib-multiline-display-paragraph)
+                 (function :tag "Custom Function")))
+
+(defun ebib-multiline-display-as-is (string)
+  "Reduce the multiline text STRING.
+The text is split into lines and returned.  No other
+modifications are made."
+  (split-string string "\n"))
+
+(defun ebib-multiline-display-paragraph (string)
+  "Reduce the multiline text STRING.
+The text is filled to account for the posibility that the
+original text is meant to be used with `visual-line-mode'.
+Return a list of strings, each a single line."
+  (split-string (with-temp-buffer
+                  (insert string)
+                  (goto-char (point-min))
+                  (forward-paragraph)
+                  (delete-region (point) (point-max))
+                  (fill-region (point-min) (point-max))
+                  (buffer-string))
+                "\n" t))
+
+(defcustom ebib-multiline-display-max-lines 10
+  "The maximum number of lines to display for multiline field values."
+  :group 'ebib
+  :type 'integer)
+
 (defcustom ebib-sort-order nil
   "The fields on which the BibTeX entries are to be sorted in the BibTeX file.
 This option is described in the manual/info file in the section
@@ -846,7 +898,7 @@ Currently, the following problems are marked:
 (defvar-local ebib--multiline-info nil "Information about the multiline text being edited.")
 (defvar ebib--log-error nil "Indicates whether an error was logged.")
 (defvar-local ebib--local-bibtex-filenames nil "A list of a buffer's .bib file(s)")
-(put 'ebib--local-bibtex-filenames 'safe-local-variable (lambda (v) (null (--remove (stringp it) v))))
+(put 'ebib--local-bibtex-filenames 'safe-local-variable (lambda (v) (null (seq-remove #'stringp v))))
 
 ;; The databases.
 
@@ -1226,7 +1278,7 @@ ignored."
   (if (stringp string)
       (string-match-p "\n" string)))
 
-(defsubst ebib--first-line (string)
+(defun ebib--first-line (string)
   "Return the first line of a multiline STRING."
   (car (split-string string "\n")))
 
@@ -1261,8 +1313,10 @@ form (\"<variable>\" \"<value>\"). If STR is not a local variable
 block, the return value is nil."
   (let ((vars (split-string str "[{}\n]+" t "[{} \t\r]+")))
     (when (and (string= (car vars) "Local Variables:")
-               (string= (-last-item vars) "End:"))
-      (--map (split-string it ": " t "[ \t]+") (-slice vars 1 -1)))))
+               (string= (car (last vars)) "End:"))
+      (mapcar (lambda (elt)
+                (split-string elt ": " t "[ \t]+"))
+              (seq-subseq vars 1 -1)))))
 
 (defun ebib--get-dialect (db)
   "Get the dialect of DB.
@@ -1278,10 +1332,10 @@ DIALECT must be a symbol, possible values are listed in
 `bibtex-dialect-list'.  If OVERWRITE is non-nil, overwrite an
 existing dialect variable, otherwise do nothing.  The return
 value is the (un)modified list."
-  (let ((ind (--find-index (string= (car it) "bibtex-dialect") vars)))
-    (if ind
+  (let ((item (seq-find (lambda (elt) (string= (car elt) "bibtex-dialect")) vars)))
+    (if item
         (when overwrite
-          (setq vars (-replace-at ind (list "bibtex-dialect" (symbol-name dialect)) vars)))
+          (setcdr item (list (symbol-name dialect))))
       (setq vars (push (list "bibtex-dialect" (symbol-name dialect)) vars)))
     vars))
 
@@ -1289,7 +1343,9 @@ value is the (un)modified list."
   "Delete the dialect definition from VARS.
 VARS is a list as returned by `ebib--local-vars-to-list'.  VARS is
 not modified, instead the new list is returned."
-  (--remove (string= (car it) "bibtex-dialect") vars))
+  (seq-remove (lambda (elt)
+                (string= (car elt) "bibtex-dialect"))
+              vars))
 
 ;; The numeric prefix argument is 1 if the user gave no prefix argument at all.
 ;; The raw prefix argument is not always a number.  So we need to do our own
@@ -1301,6 +1357,73 @@ argument to a function or not."
   (when (numberp num)
     num))
 
+;;; Functions for database access
+;;
+;; We can mostly use the ebib-db-* functions directly, but we don't want to handle
+;; the braces in `ebib-db.el', so we define some extra access functions here.
+
+(defun ebib--store-entry (entry-key fields db &optional timestamp if-exists)
+  "Store the entry defined by ENTRY-KEY and FIELDS into DB.
+Optional argument TIMESTAMP indicates whether a timestamp is to
+be added to the entry.  Note that for a timestamp to be added,
+`ebib-use-timestamp' must also be set to T. IF-EXISTS is as for
+`ebib-db-set-entry'.
+
+Return ENTRY-KEY if storing the entry was succesful, nil
+otherwise.  Depending on the value of IF-EXISTS, storing an entry
+may also result in an error."
+  (let ((result (ebib-db-set-entry entry-key fields db if-exists)))
+    (when (and result timestamp ebib-use-timestamp)
+      (ebib-set-field-value "timestamp" (format-time-string ebib-timestamp-format) entry-key db 'overwrite))
+    result))
+
+(defun ebib-set-field-value (field value key db &optional if-exists nobrace)
+  "Set FIELD to VALUE in entry KEY in database DB.
+
+IF-EXISTS determines what to do if the field already exists.  If
+it is `overwrite', the existing value is overwritten.  If it is
+`noerror', the value is not stored and the function returns nil.
+If it is nil (or any other value), an error is raised.
+
+IF-EXISTS can also be the symbol `append' or a string.  In this
+case, the new value is appended to the old value, separated by a
+space or by the string.  Before appending, braces/double quotes
+are removed from both values.  Whether braces are added to the
+new value depends on the value of NOBRACE.
+
+If NOBRACE is t, the value is stored without braces.  If it is
+nil, braces are added if not already present.  NOBRACE may also be
+the symbol `as-is', in which case the value is stored as is.
+
+A field can be removed from the entry by passing nil as VALUE and
+setting IF-EXISTS to 'overwrite.
+
+Return t upon success, or nil if the value could not be stored."
+  (if (eq if-exists 'append)
+      (setq if-exists " "))
+  (let ((old-value (ebib-db-get-field-value field key db 'norerror)))
+    ;; If the field has a value, decide what to do:
+    (if old-value
+        (cond
+         ((eq if-exists 'overwrite)
+          (setq old-value nil))
+         ((stringp if-exists)
+          (setq value (concat (ebib-unbrace old-value) if-exists (ebib-unbrace value)))
+          (setq old-value nil))
+         ((not (eq if-exists 'noerror))
+          (error "[Ebib] Field `%s' exists in entry `%s'; cannot overwrite" field key))))
+    ;; If there is (still) an old value, do nothing.
+    (unless old-value
+      ;; Otherwise overwrite the existing entry. Note that to delete a field, we
+      ;; set its value to nil, rather than removing it altogether from the
+      ;; database. In `ebib--display-fields', such fields are ignored, and they're
+      ;; not saved.
+      (if (and value nobrace)
+          (unless (eq nobrace 'as-is)
+            (setq value (ebib-unbrace value)))
+        (setq value (ebib-brace value)))
+      (ebib-db-set-field-value field value key db 'overwrite))))
+
 (defun ebib-get-field-value (field key db &optional noerror unbraced xref)
   "Return the value of FIELD in entry KEY in database DB.
 If FIELD or KEY does not exist, trigger an error, unless NOERROR
@@ -1321,7 +1444,7 @@ string has the text property `ebib--alias' with value t."
          (xref-key)
          (alias))
     (when (and (not value) xref)      ; Check if there's a cross-reference.
-      (setq xref-key (ebib-db-unbrace (ebib-db-get-field-value "crossref" key db 'noerror)))
+      (setq xref-key (ebib-unbrace (ebib-db-get-field-value "crossref" key db 'noerror)))
       (when xref-key
         (let ((xref-db (ebib--find-db-for-key xref-key db)))
           (when xref-db
@@ -1338,7 +1461,7 @@ string has the text property `ebib--alias' with value t."
     (unless (= 0 (length value)) ; (length value) == 0 if value is nil or if value is "".
       (setq value (copy-sequence value)) ; Copy the value so we can add text properties.
       (when unbraced
-        (setq value (ebib-db-unbrace value)))
+        (setq value (ebib-unbrace value)))
       (when alias
         (add-text-properties 0 1 '(ebib--alias t) value))
       (when xref
@@ -1348,6 +1471,133 @@ string has the text property `ebib--alias' with value t."
       (setq value noerror))
     value))
 
+;; Functions for writing out Bib(La)TeX entries.
+(defun ebib--format-entry (key db &optional timestamp sort)
+  "Write entry KEY in DB into the current buffer in BibTeX format.
+If TIMESTAMP is non-nil and `ebib-use-timestamp' is set, a
+timestamp is added to the entry, possibly overwriting an existing
+timestamp.  If SORT is non-nil, the fields are sorted before
+formatting the entry."
+  (let* ((entry (copy-alist (ebib-db-get-entry key db 'noerror)))
+         (type (cdr (assoc "=type=" entry))))
+    (when entry
+      (if (and timestamp ebib-use-timestamp)
+          (setcdr (assoc-string "timestamp" entry 'case-fold) (format-time-string (concat "{" ebib-timestamp-format "}"))))
+      (setq entry (seq-filter (lambda (field)
+                                (and (cdr field) ; Remove fields with value nil. See `ebib-set-field-value'.
+                                     (not (ebib--special-field-p (car field)))))
+                              entry))
+      (setq entry (if sort
+                      (cl-sort entry #'string< :key #'car)
+                    ;; When reading, fields are stored with `push', so if we don't
+                    ;; sort, we need to reverse them to get the original order
+                    ;; back.  See github issues #42, #55, #62.
+                    (reverse entry)))
+      (insert (format "@%s{%s,\n" type key))
+      (insert (mapconcat (lambda (field)
+                           (format "\t%s = %s" (car field) (cdr field)))
+                         entry
+                         ",\n"))
+      (insert "\n}\n\n"))))
+
+(defun ebib--format-comments (db)
+  "Write the @COMMENTS of DB into the current buffer in BibTeX format."
+  (mapc (lambda (c)
+          (insert (format "@Comment%s\n\n" c)))
+        (ebib-db-get-comments db)))
+
+(defun ebib--format-strings (db)
+  "Write the @Strings of DB into the current buffer in BibTeX format."
+  (mapc (lambda (str)
+          (insert (format "@String{%s = %s}\n" (car str) (cdr str))))
+        (ebib-db-get-all-strings db))
+  (insert "\n"))
+
+(defun ebib--get-sortstring (entry-key sortkey-list db)
+  "Return the field value on which the entry ENTRY-KEY is to be sorted.
+SORTKEY-LIST is a list of fields that are considered in order for
+the sort value.  DB is the database that contains the entry
+referred to by ENTRY-KEY."
+  (let ((sort-string nil))
+    (while (and sortkey-list
+                (null (setq sort-string (ebib-get-field-value (car sortkey-list) entry-key db 'noerror 'unbraced))))
+      (setq sortkey-list (cdr sortkey-list)))
+    sort-string))
+
+(defun ebib--format-local-vars (db)
+  "Write the local variables of DB into the current buffer."
+  (let ((lvars (ebib-db-get-local-vars db)))
+    (when lvars
+      (insert (concat "@Comment{\n"
+                      ebib-local-variable-indentation "Local Variables:\n"
+                      (mapconcat (lambda (e) (format "%s%s: %s\n" ebib-local-variable-indentation (car e) (cadr e))) lvars "")
+                      ebib-local-variable-indentation "End:\n"
+                      "}\n\n")))))
+
+(defun ebib--format-database-as-bibtex (db)
+  "Write database DB into the current buffer in BibTeX format."
+  (ebib--format-comments db)
+  (when (ebib-db-get-preamble db)
+    (insert (format "@Preamble{%s}\n\n" (ebib-db-get-preamble db))))
+  (ebib--format-strings db)
+  ;; We define two comparison functions for `sort'.  These must simply
+  ;; return non-NIL if the first element is to be sorted before the second.
+  (cl-flet
+      ;; The first one is simple: if X has a crossref field, it must be
+      ;; sorted before Y (or at least *can* be, if Y also has a crossref
+      ;; field).
+      ((compare-xrefs (x _)
+                      (ebib-get-field-value "crossref" x db 'noerror))
+       ;; This one's a bit trickier.  We iterate over the lists of fields in
+       ;; `ebib-sort-order'.  For each level, `ebib--get-sortstring' then returns the
+       ;; string that can be used for sorting.  If all fails, sorting is done on
+       ;; the basis of the entry key.
+       (entry< (x y)
+               (let (sortstring-x sortstring-y)
+                 (cl-loop for sort-list in ebib-sort-order do
+                          (setq sortstring-x (ebib--get-sortstring x sort-list db))
+                          (setq sortstring-y (ebib--get-sortstring y sort-list db))
+                          while (cl-equalp sortstring-x sortstring-y))
+                 (if (and sortstring-x sortstring-y)
+                     (string< sortstring-x sortstring-y)
+                   (string< x y)))))    ; compare entry keys
+    ;; Only entries in visible the index buffer are saved, in case we're writing
+    ;; a filtered db to a new file.
+    (let ((sorted-list (sort (ebib-db-list-keys ebib--cur-db) #'string<)))
+      (cond
+       (ebib-save-xrefs-first
+        (setq sorted-list (sort sorted-list #'compare-xrefs)))
+       (ebib-sort-order
+        (setq sorted-list (sort sorted-list #'entry<))))
+      (mapc (lambda (key) (ebib--format-entry key db nil)) sorted-list))
+    (ebib--format-local-vars db)))
+
+(defun ebib--export-entries-to-db (entries target-db source-db)
+  "Export ENTRIES from SOURCE-DB to TARGET-DB.
+ENTRIES is a list of entry keys."
+  (let ((target-keys-list (ebib-db-list-keys target-db))
+        (modified nil))
+    (mapc (lambda (key)
+            (if (member key target-keys-list)
+                (ebib--log 'message "Entry key `%s' already exists in database %s" key (ebib-db-get-filename target-db 'short))
+              (ebib--store-entry key (copy-tree (ebib-db-get-entry key source-db)) target-db t)
+              (setq modified t)))
+          entries)
+    (when modified
+      (ebib-db-kill-buffer target-db)
+      (ebib-db-set-modified t target-db))))
+
+(defun ebib--export-entries-to-file (entries filename source-db)
+  "Export ENTRIES from SOURCE-DB to FILENAME.
+ENTRIES is a list of entry keys."
+  (with-temp-buffer
+    (insert "\n")
+    (mapc (lambda (key)
+            (ebib--format-entry key source-db nil))
+          entries)
+    (append-to-file (point-min) (point-max) filename)
+    (setq ebib--export-filename filename)))
+
 (defun ebib--find-db-for-key (key db)
   "Return the database containing KEY.
 First see if KEY is in DB.  If not, check all members of
@@ -1356,7 +1606,7 @@ nil."
   (if (ebib-db-get-entry key db 'noerror)
       db
     (let ((dbs (reverse ebib--databases)))
-      (while (and dbs (not (ebib-db-get-entry key (car dbs) 'noerror)))
+      (while (and dbs (not (ebib-db-has-key key (car dbs))))
         (setq dbs (cdr dbs)))
       (car dbs))))
 
@@ -1382,6 +1632,109 @@ inherit a value, this function returns nil."
       (or (car (rassoc (downcase target-field) inheritance))
           target-field))))
 
+;; ebib-set-string is just a front-end for `ebib-db-set-string'.  The only additional
+;; functionality is that it wraps the string value in braces if it isn't
+;; already.  We need to do this because the brace functions are in this file and
+;; `ebib-db.el' cannot depend on `ebib-utils.el' (because `ebib-utils.el' already depends on
+;; `ebib-db.el').  Similar considerations apply to `ebib-get-string' below.
+
+(defun ebib-set-string (abbr value db &optional overwrite)
+  "Set the @string definition ABBR to VALUE in database DB.
+If ABBR does not exist, create it.  OVERWRITE functions as in
+`ebib-db-set-string'.  VALUE is enclosed in braces if it isn't
+already.
+
+This function basically just calls `ebib-db-set-string' to do the
+  real work."
+  (ebib-db-set-string abbr (if (ebib-unbraced-p value)
+                           (ebib-brace value)
+                         value)
+                  db overwrite))
+
+(defun ebib-get-string (abbr db &optional noerror unbraced)
+  "Return the value of @STRING definition ABBR in database DB.
+NOERROR functions as in `ebib-db-get-string', which this
+functions calls to get the actual value.  The braces around the
+value are removed if UNBRACED is non-nil."
+  (let ((value (ebib-db-get-string abbr db noerror)))
+    (if unbraced
+        (ebib-unbrace value)
+      value)))
+
+;; EBIB-UNBRACED-P determines if STRING is enclosed in braces.  Note that we
+;; cannot do this by simply checking whether STRING begins with { and ends with
+;; } (or begins and ends with "), because something like "{abc} # D # {efg}"
+;; would then be incorrectly recognised as braced.  So we need to do the
+;; following: take out everything that is between braces or quotes, and see if
+;; anything is left.  If there is, the original string was braced, otherwise it
+;; was not.
+
+;; So we first check whether the string begins with { or ".  If not, we
+;; certainly have an unbraced string.  (EBIB-UNBRACED-P recognises this
+;; through the default clause of the COND.)  If the first character is { or ",
+;; we first take out every occurrence of backslash-escaped { and } or ", so that
+;; the rest of the function does not get confused over them.
+
+;; Then, if the first character is {, REMOVE-FROM-STRING takes out every
+;; occurrence of the regex "{[^{]*?}", which translates to "the smallest string
+;; that starts with { and ends with }, and does not contain another {.  IOW, it
+;; takes out the innermost braces and their contents.  Because braces may be
+;; embedded, we have to repeat this step until no more balanced braces are found
+;; in the string.  (Note that it would be unwise to check for just the occurrence
+;; of { or }, because that would throw EBIB-UNBRACED-P in an infinite loop if
+;; a string contains an unbalanced brace.)
+
+;; For strings beginning with " we do the same, except that it is not
+;; necessary to repeat this in a WHILE loop, for the simple reason that
+;; strings surrounded with double quotes cannot be embedded; i.e.,
+;; "ab"cd"ef" is not a valid (BibTeX) string, while {ab{cd}ef} is.
+
+;; Note: because these strings are to be fed to BibTeX and ultimately
+;; (La)TeX, it might seem that we don't need to worry about strings
+;; containing unbalanced braces, because (La)TeX would choke on them.  But
+;; the user may inadvertently enter such a string, and we therefore need to
+;; be able to handle it.  (Alternatively, we could perform a check on
+;; strings and warn the user.)
+
+(defun ebib-unbraced-p (string)
+  "Non-nil if STRING is not enclosed in braces or quotes."
+  (cl-flet ((remove-from-string (string remove)
+                                (apply #'concat (split-string string remove))))
+    (when (stringp string)
+      (cond
+       ((eq (string-to-char string) ?\{)
+        ;; first, remove all escaped { and } from the string:
+        (setq string (remove-from-string (remove-from-string string "[\\][{]")
+                                         "[\\][}]"))
+        ;; then remove the innermost braces with their contents and continue until
+        ;; no more braces are left.
+        (while (and (member ?\{ (string-to-list string)) (member ?\} (string-to-list string)))
+          (setq string (remove-from-string string "{[^{]*?}")))
+        ;; if STRING is not empty, the original string contains material not in braces
+        (> (length string) 0))
+       ((eq (string-to-char string) ?\")
+        ;; remove escaped ", then remove any occurrences of balanced quotes with
+        ;; their contents and check for the length of the remaining string.
+        (> (length (remove-from-string (remove-from-string string "[\\][\"]")
+                                       "\"[^\"]*?\""))
+           0))
+       (t t)))))
+
+(defun ebib-unbrace (string)
+  "Convert STRING to its unbraced counterpart.
+If STRING is already unbraced, do nothing."
+  (if (and (stringp string)
+           (not (ebib-unbraced-p string)))
+      (substring string 1 -1)
+    string))
+
+(defun ebib-brace (string)
+  "Put braces around STRING.
+If STRING is already braced, do nothing."
+  (if (ebib-unbraced-p string)
+      (concat "{" string "}")
+    string))
+
 (defun ebib--list-fields (entry-type type dialect)
   "List the fields of ENTRY-TYPE.
 TYPE specifies which fields to list.  It is a symbol and can be
@@ -1407,7 +1760,9 @@ which ENTRY-TYPE is an alias."
     (when (memq type '(optional extra all))
       (setq optional (mapcar #'car (nth 4 (assoc-string entry-type (bibtex-entry-alist dialect) 'case-fold)))))
     (when (memq type '(all extra))
-      (setq extra (--remove (member-ignore-case it (append required optional)) (cdr (assq dialect ebib-extra-fields)))))
+      (setq extra (seq-remove (lambda (elt)
+                                (member-ignore-case elt (append required optional)))
+                              (cdr (assq dialect ebib-extra-fields)))))
     (cond
      ((eq type 'required) required)
      ((eq type 'optional) optional)
@@ -1426,7 +1781,9 @@ in `bibtex-dialect-list' or NIL, in which case the value of
 `ebib-bibtex-dialect' is used."
   (or dialect (setq dialect ebib-bibtex-dialect))
   (let ((fields (ebib--list-fields (cdr (assoc "=type=" entry)) 'all dialect)))
-    (--remove (member-ignore-case (car it) (cons "=type=" fields)) entry)))
+    (seq-remove (lambda (elt)
+                  (member-ignore-case (car elt) (cons "=type=" fields)))
+                entry)))
 
 (defun ebib--list-entry-types (&optional dialect include-aliases)
   "Return a list of entry types.
@@ -1452,7 +1809,7 @@ Possible values for DIALECT are those listed in
   (or (cdr (assq dialect ebib--unique-field-alist))
       (let (fields)
         (mapc (lambda (entry)
-                (setq fields (-union fields (ebib--list-fields (car entry) 'all dialect))))
+                (setq fields (seq-uniq (seq-concatenate 'list fields (ebib--list-fields (car entry) 'all dialect)))))
               (bibtex-entry-alist dialect))
         (push (cons dialect fields) ebib--unique-field-alist)
         fields)))
@@ -1468,15 +1825,18 @@ Otherwise, If FIELD is \"Entry Key\", KEY is returned; if FIELD
 is \"Author/Editor\", the contents of the Author field is
 returned or, if the Author field is empty, the contents of the
 Editor field; if the Editor field is empty as well, return the
-string \"(No Author/Editor)\".  These defaults can be overridden
-by adding an appropriate transformation function to
+string \"(No Author/Editor)\".  If FIELD is \"Year\", the year of
+the first date in the Date field is returned (i.e., date ranges
+are ignored), or, if no year can be found in the Date field, the
+contents of the Year field is returned.  If this field is empty,
+the string \"XXXX\" is returned.  These defaults can be
+overridden by adding an appropriate transformation function to
 `ebib-field-transformation-functions'.
 
 If FIELD is not found in `ebib-field-transformation-functions'
 and is not one of the special fields listed above, the field
 value is returned.  If a field value is empty, the return value
-is the string \"(No <FIELD>)\", or, if FIELD is \"Year\", the
-string \"(XXXX)\"."
+is the string \"(No <FIELD>)\"."
   (cond
    ((assoc-string field ebib-field-transformation-functions 'case-fold)
     (funcall (cdr (assoc field ebib-field-transformation-functions)) field key db))
@@ -1486,9 +1846,24 @@ string \"(XXXX)\"."
     (or (ebib-get-field-value "Author" key db 'noerror 'unbraced 'xref)
         (ebib-get-field-value "Editor" key db "(No Author/Editor)" 'unbraced 'xref)))
    ((cl-equalp field "Year")
-    (ebib-get-field-value "Year" key db "XXXX" 'unbraced 'xref))
+    (ebib-get-year-for-display key db))
    (t (ebib-get-field-value field key db (format "(No %s)" (capitalize field)) 'unbraced 'xref))))
 
+(defun ebib-get-year-for-display (key db)
+  "Return the year for entry KEY in DB.
+In BibTeX, this simply returns the value of the Year field, or
+the string \"XXXX\" if this is empty.  For BibLaTeX, this first
+checks the Date field and returns the year from the first date in
+this field (meaning date ranges are ignored).  If no year can be
+found in the Date field, the value of the Year field is returned,
+or \"XXXX\" if it's empty."
+  (save-match-data
+    (let ((date (ebib-get-field-value "Date" key db 'noerror 'unbraced 'xref)))
+      (if (and date
+               (string-match "^\\(?:\\.\\.\\)?/?\\([0-9]\\{4\\}\\)" date))
+          (match-string 1 date)
+        (ebib-get-field-value "Year" key db "XXXX" 'unbraced 'xref)))))
+
 (defun ebib-clean-TeX-markup (field key db)
   "Return the contents of FIELD from KEY in DB without TeX markups."
   (let ((str (ebib-get-field-value field key db "" 'unbraced 'xref)))
@@ -1522,15 +1897,6 @@ are returned as a string."
                         '("A" "An" "At" "Of" "On" "And" "For"))))
       str)))
 
-(defun ebib-display-note-symbol (_field key _db)
-  "Return the note symbol for displaying if a note exists for KEY."
-  (if (ebib--notes-exists-note key)
-      (propertize ebib-notes-symbol
-                  'face '(:height 0.8 :inherit link)
-                  'mouse-face 'highlight)
-    (propertize (make-string (string-width ebib-notes-symbol) ?\s)
-                'face '(:height 0.8))))
-
 (defun ebib-display-www-link (field key db)
   "Return the content of FIELD from KEY in DB as a link.
 This function is mainly intended for the DOI and URL fields."
@@ -1558,9 +1924,11 @@ characters in fields."
   ;; First sort the keys themselves.
   (setq keys (sort keys #'string<))
   ;; And then stably sort on the sort field.
-  (let* ((field (if (ebib-db-custom-sorted-p db)
-                    (ebib-db-get-sort-field db)
-                  (caar ebib-index-columns)))
+  (let* ((sortinfo (or (ebib-db-get-sortinfo db)
+                       ebib-index-default-sort
+                       (cons (caar ebib-index-columns) 'ascend)))
+         (field (car sortinfo))
+         (direction (cdr sortinfo))
          ;; We use a temp list for sorting, so that the :key argument to
          ;; `cl-stable-sort' can simply be `car' rather than (a much
          ;; heavier) `ebib-get-field-value'. Sorting is much faster
@@ -1568,14 +1936,11 @@ characters in fields."
          (list (mapcar (lambda (key)
                          (cons (ebib--get-field-value-for-display field key db) key))
                        keys)))
-    (setq list (cl-stable-sort list (if (fboundp 'string-collate-lessp)
-                                        #'string-collate-lessp
-                                      #'string-lessp)
-                               :key #'car))
-    (setq keys (mapcar #'cdr list)))
-  ;; Reverse the list if necessary.
-  (if (eq (ebib-db-get-sort-order db) 'descend)
-      (setq keys (nreverse keys)))
+    (setq list (cl-stable-sort list #'string-collate-lessp :key #'car))
+    (setq keys (mapcar #'cdr list))
+    ;; Reverse the list if necessary.
+    (if (eq direction 'descend)
+        (setq keys (nreverse keys))))
   ;; Now return the list of keys.
   keys)
 
